# Script to execute CASA-based "polsimulate"
# (c) I. Marti-Vidal (Univ. Valencia) 2020.
############
# Script to simulate full-polarization EHT observations of a 
# resolved calibrator.
# This script complements the contents of the paper:

# I. Marti-Vidal, A. Mus, P. de Vicente & J. Gonzalez (2020)


############
### ADAPTED FOR CASA-VLBI WORKSHOP/TUTORIAL 2023
############




#########################
### CONFIGURATION HEADER
#########################

## CASES:
##
## 0 -> Unpolarized point source. No Dterms
## 1 -> Unpolarized point source. With Dterms
## 2 -> Polarized point source. No Dterms
## 3 -> Polarized point source With Dterms
## 33 -> Polarized point source with Dterms for PolCal
##
## 4 -> Polarized double source. No Dterms.
## 5 -> Polarized double source. With Dterms.
## 6 -> Polarized double source (changed polarization). With Dterms.
## 7 -> Polarized double source. With freq-dependent Dterms.
##      (Dterms will run from a factor 0.5 (nu=0) to 1.0 (nu=NCHAN)


## Which case do we simulate (see list above):
CASE = 5


#########################
### END OF CONFIGURATION HEADER
#########################












import numpy as np
import os
import pickle as pk
from task_polsimulate import polsimulate as polsimulate

# Set random seed:
np.random.seed(42)





## Name of the measurement set:
VISNAME = 'POLSIM_CASE-%i'%CASE


## Number of frequency channels:
NCHAN = 1


# True means to CLEAN (manually) all four Stokes parameters:
# The mask made from this step will be used later in the 
# "polarization self-calibration" steps:
DOCLEAN = True



## DTERMS (IN THE CASES THEY ARE APPLIED; see list above):
# Get random Dterms from a Gaussian distribution:
DRs = (np.random.normal(0.,0.20,8)+np.random.normal(0.,0.20,8)*1.j)
DLs = (np.random.normal(0.,0.20,8)+np.random.normal(0.,0.20,8)*1.j)






#####################
### SCRIPT STARTS ###
#####################


## Set configuration parameters for the simulator:

muas = 1./3600.*1.e-6

#########
## POINT SOURCES:

if CASE in [0,1]:
  I = [1.0]; Qfrac = [0.0]; Ufrac = [0.0]; Vfrac = [0.0]; 
  RM = [0.0]; spec_index = [0.0]
  RAoffset = [0.0] ; Decoffset = [0.0]
  if CASE == 0:
    DRs[:] = 0.0  ; DLs[:] = 0.0

elif CASE in [2,3,33]:
  I = [1.0]; Qfrac = [0.1]; Ufrac = [0.0]; Vfrac = [0.02]; 
  RM = [0.0]; spec_index = [0.0]
  RAoffset = [0.0] ; Decoffset = [0.0]
  if CASE == 2:
    DRs[:] = 0.0  ; DLs[:] = 0.0


#########
## DOUBLE SOURCE 1:

elif CASE in [4,5]:
  I = [1.0, 0.75]; Qfrac = [0.1, 0.0]; Ufrac = [0.0, 0.05]; Vfrac = [0.02, -0.01]; 
  RM = [0.0, 0.0]; spec_index = [0.0, 0.0]
  RAoffset = [0.0, 50.*muas] ; Decoffset = [0.0, 20.*muas]
  if CASE == 4:
    DRs[:] = 0.0  ; DLs[:] = 0.0


#########
## DOUBLE SOURCE 2:

elif CASE in [6,7]:
  I = [1.0, 0.75]; Qfrac = [0.0, 0.05]; Ufrac = [-0.1, 0.0]; Vfrac = [-0.02, 0.01]; 
  RM = [0.0, 0.0]; spec_index = [0.0, 0.0]
  RAoffset = [0.0, 50.*muas] ; Decoffset = [0.0, 20.*muas]
  if CASE == 7:
    DRnu = np.zeros((len(DRs),NCHAN),dtype=np.complex128)
    DLnu = np.zeros((len(DLs),NCHAN),dtype=np.complex128)
    DRnu[:,:] = DRs[:,np.newaxis];  DLnu[:,:] = DLs[:,np.newaxis]; 
    DRnu *= np.linspace(0.5,1.,NCHAN)
    DLnu *= np.linspace(0.5,1.,NCHAN)
    DRs = DRnu; DLs = DLnu

else:
  raise Exception("Unknown CASE id. %i"%(int(CASE)))






######################################
### THIS IS FOR THE POLCAL CASE:
if CASE in [33]:

## ANTENNA MOUNTS: All antennas have ALT-AZ mounts:
## If you use a list, beware of the antenna order in the "EHT.cfg" file!
  # mounts  =  ['AZ', 'AZ', 'AZ', 'AZ', 'AZ', 'AZ', 'AZ', 'AZ']

# Otherwise, an empty dictionary will set them all to alt-az:
  mounts = {}

else:

## True EHT antenna mounts:
## If you use a list, beware of the antenna order in the "EHT.cfg" file!
#  mounts  =  ['AZ', 'NR', 'NR', 'AZ', 'NL', 'NL', 'AZ', 'NL']

# Otherwise, the dictionary only needs to have the antennas with funny mounts:
  mounts = {'AP':'NR',  'AZ':'NR',  'LM':'NL', 'SM':'NL', 'PV':'NL'}
#######################################






# (Re-)generate measurement set:
os.system('rm -rf %s.*'%VISNAME)



## Generate the data with parallactic angle applied:

polsimulate(vis='%s.ms'%VISNAME,
  elevation_cutoff = 15.0,
  mounts=mounts, apply_parang = True,
  reuse=False, # A new Measurement Set will be made from scratch.
  model_Dt_0 = DRs, # Dterms in first polarizer (i.e., R)
  model_Dt_1 = DLs, # Dterms in second polarizer (i.e., L)
  feed = 'circular', # The polarizers are actually R and L.
  I = I, # Stokes I of the 5 source components.
  Q_frac = Qfrac, # Fractional Stokes Q.
  U_frac = Ufrac, # Fractional Stokes U.
  V_frac = Vfrac, # Fractional V
  RM = RM,  # Rotation measures.
  spec_index = spec_index, # Spectral indices.
## RA offsets of each of the five source components (given in degrees):
  RA_offset = RAoffset,
## Samce for the Dec offsets (all are zero, since the jet is in E-W direction:
  Dec_offset = Decoffset,
## Phase center of the source core (that of SgrA*)
  phase_center = "J2000 17h45m40.4230 -29d00m28.0400",
  array_configuration    =  "EHT.cfg", # ascii file with the antenna info.
  LO = 2.3e+11, BBs = [0.0], spw_width = 1.e9, nchan = NCHAN, # Freq. config.
  nscan = "TRACK_C.listobs", # Take observing times from this listobs file.
  visib_time = '10s', # VLBI integration time.
  export_uvf = False, # Export in uvfits.
  corrupt=True) # Add noise using the "sm" tool algorithm.



## If polcal is going to be used, generate the data without parangle correction:
if CASE in [33]:
  os.system('rm -rf %s_for_PolCal.*'%VISNAME)
  polsimulate(vis='%s_for_PolCal.ms'%VISNAME,
    mounts=mounts, apply_parang = False,
    reuse=False, # A new Measurement Set will be made from scratch.
    model_Dt_0 = DRs, # Dterms in first polarizer (i.e., R)
    model_Dt_1 = DLs, # Dterms in second polarizer (i.e., L)
    feed = 'circular', # The polarizers are actually R and L.
    I = I, # Stokes I of the 5 source components.
    Q_frac = Qfrac, # Fractional Stokes Q.
    U_frac = Ufrac, # Fractional Stokes U.
    V_frac = Vfrac, # Fractional V
    RM = RM,  # Rotation measures.
    spec_index = spec_index, # Spectral indices.
## RA offsets of each of the five source components (given in degrees):
    RA_offset = RAoffset,
## Samce for the Dec offsets (all are zero, since the jet is in E-W direction:
    Dec_offset = Decoffset,
## Phase center of the source core (that of SgrA*)
    phase_center = "J2000 17h45m40.4230 -29d00m28.0400",
    array_configuration    =  "EHT.cfg", # ascii file with the antenna info.
    LO = 2.3e+11, BBs = [0.0], spw_width = 1.e9, nchan = NCHAN, # Freq. config.
    nscan = "TRACK_C.listobs", # Take observing times from this listobs file.
    visib_time = '10s', # VLBI integration time.
    export_uvf = False, # Export in uvfits.
    corrupt=True) # Add noise using the "sm" tool algorithm.





### SAVE THE SIMULATED DTERMS (TO ASSESS THE PERFORMANCE OF POLSOLVE):
# Write the simulated Dterms values into an external file:
outp = open('%s_Dterms.dat'%VISNAME,'wb')
pk.dump([DRs,DLs], outp)
outp.close()

 
